"""Client to interact with The Movie DB (TMDB)."""
from typing import Any
from typing import Dict
from typing import List

from requests import Session


class TmdbClient:
    """Class to abstract to the Tmdb API."""

    def __init__(self, api_key: str) -> None:
        """Initialize the client.

        Args:
            api_key: an api_key for authentication
        """
        super().__init__()
        self.session = Session()
        self.api_key = api_key
        self._base = "https://api.themoviedb.org/3"

    def search_movies(self, query: str, region: str = "us") -> List[Dict[str, Any]]:
        """Search for movies.

        Args:
            query: the search query
            region: the region for the query, affects the release date value

        Returns:
            List[Dict[str, Any]]: the search results
        """
        payload = {
            "api_key": self.api_key,
            "language": "en-US",
            "query": query,
            "include_adult": "false",
            "region": region.upper(),
        }
        res = self.session.get(f"{self._base}/search/movie", params=payload)

        res.raise_for_status()

        results: List[Dict[str, Any]] = res.json()["results"]
        return results

    def get_streaming_providers(
        self, movie_id: str, regions: List[str]
    ) -> Dict[str, Any]:
        """Return a list of streaming providers for a given movie.

        Args:
            movie_id: the tmdb id of the movie
            regions: a list of regions to trim down the return list

        Returns:
            Dict[str, Any]: a dictionary of streaming providers, keyed by region name
        """
        payload = {"api_key": self.api_key}
        res = self.session.get(
            f"{self._base}/movie/{movie_id}/watch/providers", params=payload
        )

        res.raise_for_status()

        results: Dict[str, Any] = res.json()["results"]

        return {key: results[key.upper()] for key in regions}



"""Module to hold `phylm` tools."""
import os
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from imdb.Movie import Movie

from phylm.clients.tmdb import TmdbClient
from phylm.errors import NoTMDbApiKeyError
from phylm.sources.imdb import ia


def search_movies(query: str) -> List[Dict[str, Union[str, int]]]:
    """Return a list of search results for a query.

    Args:
        query: the search query

    Returns:
        a list of search results
    """
    results: List[Movie] = ia.search_movie(query)

    return [
        {
            "title": r.data.get("title"),
            "kind": r.data.get("kind"),
            "year": r.data.get("year"),
            "imdb_id": r.movieID,
            "cover_photo": r.data.get("cover url"),
        }
        for r in results
    ]


def _initialize_tmdb_client(api_key: Optional[str] = None) -> TmdbClient:
    """Initialize and return a TmdbClient.

    Args:
        api_key: an optional api_key to take precedence over an env var key

    Raises:
        NoTMDbApiKeyError: when no api_key has been provided

    Returns:
        TmdbClient: an authorized Tmdb client
    """
    tmdb_api_key = api_key or os.environ.get("TMDB_API_KEY")

    if not tmdb_api_key:
        raise NoTMDbApiKeyError("An `api_key` must be provided to use this service")

    return TmdbClient(api_key=tmdb_api_key)


def search_tmdb_movies(
    query: str, api_key: Optional[str] = None, region: str = "us"
) -> List[Dict[str, Any]]:
    """Search for movies on TMDb.

    Args:
        query: the query string
        api_key: an api_key can either be provided here or through a TMDB_API_KEY env
            var
        region: an optional region to provide with the search request, affects the
            release_date value returned, must be provided in ISO 3166-1 format (eg.
            "us" or "gb")

    Returns:
        List[Dict[str, Any]]: the search results
    """
    client = _initialize_tmdb_client(api_key=api_key)

    return client.search_movies(query=query, region=region)


def get_streaming_providers(
    tmdb_movie_id: str,
    regions: List[str],
    api_key: Optional[str] = None,
) -> Dict[str, Any]:
    """Return a list of streaming providers for a given movie.

    Args:
        tmdb_movie_id: the tmdb id of the movie
        regions: a list of regions to trim down the return list
        api_key: an api_key can either be provided here or through a TMDB_API_KEY env
            var

    Returns:
        Dict[str, Any]: a dictionary of streaming providers, keyed by region name
    """
    client = _initialize_tmdb_client(api_key=api_key)

    return client.get_streaming_providers(movie_id=tmdb_movie_id, regions=regions)



"""Module to define the Mtc class."""
import re
from typing import Optional

from aiohttp import ClientSession
from bs4 import BeautifulSoup
from bs4.element import Tag

from phylm.utils.web import async_soupify
from phylm.utils.web import url_encode

MTC_BASE_MOVIE_URL = "https://www.metacritic.com/search/movie"


class Mtc:
    """Class to abstract a Metacritic movie search result."""

    def __init__(self, raw_title: str, raw_year: Optional[int] = None) -> None:
        """Initialize the object.

        Args:
            raw_title: the given title of the movie
            raw_year: an optional year for improved matching
        """
        self.raw_title = raw_title
        self.raw_year = raw_year
        self.low_confidence = False
        self._mtc_data: Optional[Tag] = None

    def _parse_data(self, soup: BeautifulSoup) -> Optional[Tag]:
        results = soup.find_all("li", {"class": "result"})

        if not results:
            return None

        # first try matching on year
        for result in results:
            year = _extract_year(result)
            if self.raw_year and self.raw_year == year:
                return result

        # then try matching on title
        for result in results:
            result_title: str = result.find("a").string.strip()
            if result_title.lower().strip() == self.raw_title.lower().strip():
                return result

        # finally pick the first result
        self.low_confidence = True
        return results[0]

    async def _scrape_data(
        self, session: Optional[ClientSession] = None
    ) -> BeautifulSoup:
        url_encoded_film = url_encode(self.raw_title)
        search_url = f"{MTC_BASE_MOVIE_URL}/{url_encoded_film}/results"
        return await async_soupify(search_url, session)

    async def load_source(self, session: Optional[ClientSession] = None) -> None:
        """Asynchronously load the data from the source.

        Args:
            session: an optional instance of `aiohttp.ClientSession` in which to run the
                request
        """
        raw_data = await self._scrape_data(session=session)
        self._mtc_data = self._parse_data(raw_data)

    @property
    def title(self) -> Optional[str]:
        """Return the title.

        Returns:
            the title of the movie
        """
        if not self._mtc_data:
            return None
        title_tag = self._mtc_data.find("a")
        if title_tag:
            return str(title_tag.get_text()).strip()
        return None

    @property
    def year(self) -> Optional[int]:
        """Return the year.

        Returns:
            the year of the movie
        """
        if not self._mtc_data:
            return None
        return _extract_year(self._mtc_data)

    @property
    def rating(self) -> Optional[str]:
        """Return the rating.

        Returns:
            the mtc rating
        """
        if not self._mtc_data:
            return None
        rating_tag = self._mtc_data.find("span", {"class": "metascore_w"})
        if not rating_tag:
            return None
        return str(rating_tag.get_text().strip())


def _extract_year(tag: Tag) -> Optional[int]:
    """Return the year if it exists in a tag.

    Args:
        tag: the mtc search result bs4 tag to search

    Returns:
        Optional[int]: the year of the movie
    """
    year_tag = tag.find("p")

    if not year_tag:
        return None

    year_search = re.search(r"\d{4}", year_tag.get_text())

    if not year_search:
        return None

    return int(year_search.group())



"""Module to contain the `Phylm` class definition."""
import asyncio
from typing import List
from typing import Optional

from aiohttp import ClientSession

from phylm.errors import SourceNotLoadedError
from phylm.errors import UnrecognizedSourceError
from phylm.sources import Imdb
from phylm.sources import Mtc
from phylm.sources import Rt


class Phylm:
    """Main `Phylm` entrypoint."""

    def __init__(
        self,
        title: str,
        imdb_id: Optional[str] = None,
        year: Optional[int] = None,
    ) -> None:
        """Initialize a `Phylm` object.

        Args:
            title: the title of the movie
            imdb_id: an optional `IMDb` ID of the movie
            year: an optional year of the movie
        """
        self.title = title
        self.imdb_id = imdb_id
        self.year = year
        self._imdb: Optional[Imdb] = None
        self._mtc: Optional[Mtc] = None
        self._rt: Optional[Rt] = None

    def __repr__(self) -> str:
        """Return the string representation.

        Returns:
            the string representation
        """
        return f"<class '{self.__class__.__name__}' title:'{self.title}'>"

    def __str__(self) -> str:
        """Return the string representation."""
        return self.__repr__()

    @property
    def imdb(self) -> Imdb:
        """Return the IMDb data.

        Returns:
            The IMDb data

        Raises:
            SourceNotLoadedError: if the source is not loaded
        """
        if self._imdb is None:
            raise SourceNotLoadedError("The data for Imdb has not yet been loaded")

        return self._imdb

    @property
    def mtc(self) -> Mtc:
        """Return the Metacritic data.

        Returns:
            The Metacritic data

        Raises:
            SourceNotLoadedError: if the source is not loaded
        """
        if self._mtc is None:
            raise SourceNotLoadedError(
                "The data for Metacritic has not yet been loaded"
            )

        return self._mtc

    @property
    def rt(self) -> Rt:
        """Return the Rotten Tomatoes data.

        Returns:
            The Rotten Tomatoes data

        Raises:
            SourceNotLoadedError: if the source is not loaded
        """
        if self._rt is None:
            raise SourceNotLoadedError(
                "The data for Rotten Tomatoes has not yet been loaded"
            )

        return self._rt

    async def load_source(
        self,
        source: str,
        imdb_id: Optional[str] = None,
        session: Optional[ClientSession] = None,
    ) -> "Phylm":
        """Asynchronously load the film data for a source.

        Args:
            source: the desired source
            imdb_id: an optional `IMDb` id which will be used to load the imdb data
                instead of a basic search on the title
            session: an optional instance of `aiohttp.ClientSession` in which to run the
                request

        Returns:
            the instance

        Raises:
            UnrecognizedSourceError: if the source is not recognized
        """
        if source == "imdb":
            if not self._imdb:
                movie_id = imdb_id or self.imdb_id
                self._imdb = Imdb(
                    raw_title=self.title,
                    movie_id=movie_id,
                    raw_year=self.year,
                )
                await self._imdb.load_source()
            return self

        if source == "mtc":
            if not self._mtc:
                self._mtc = Mtc(raw_title=self.title, raw_year=self.year)
                await self._mtc.load_source(session=session)
            return self

        if source == "rt":
            if not self._rt:
                self._rt = Rt(raw_title=self.title, raw_year=self.year)
                await self._rt.load_source(session=session)
            return self

        raise UnrecognizedSourceError(f"{source} is not a recognized source")

    async def load_sources(
        self,
        sources: List[str],
    ) -> "Phylm":
        """Asynchronously load multiple sources.

        Args:
            sources: a list of the desired sources

        Returns:
            the instance
        """
        session = ClientSession()

        try:
            await asyncio.gather(
                *[self.load_source(source, session=session) for source in sources]
            )
        finally:
            await session.close()

        return self



"""Module to hold custom error definitions."""


class UnrecognizedSourceError(Exception):
    """Raised when an unrecognized source is given."""


class SourceNotLoadedError(Exception):
    """Raised when data from an unloaded source is retreived."""


class NoTMDbApiKeyError(Exception):
    """Raised when requests are made to TMDb but no api_key has be provided."""
